#include "population.h"

void Population::runGeneration(){
    eval();
    sortPopulation();
    mapFitnesses();
    logData();
    selection();
    generation++;
    z = generation%2;
    crossover();
    mutation();
}

void Population::mapFitnesses(){
    mappedFitnessStats.clear();
    for( int i = 0; i < genomes[z].Size(); i ++ ){
        mapFitness( genomes[z][i] );
        mappedFitnessStats.add( genomes[z][i].mappedFitness );
    }
}

Population::Population(){
    generation = 0;
    z = 0;
}
void Population::setParams( Params* params ){
    this->params = params;
}
void Population::initialize(){
    Genome g;
    g.setParams( params );
    g.initialize();
    for( int i = 0; i < params->popSize; i ++ ){
        g.randomize();
        genomes[0].Append( g );
        genomes[1].Append( g );
    }
}
char* Population::toString(){
    string[0] = '\0';
    for( int i = 0; i < params->popSize; i ++ ){
        strcat( string, genomes[z][i].toString().Get() );
        strcat("\n");
    }
    return string;
}
void Population::eval(){
    assert( params );
    assert( params->evaluator );
    fitnessStats.clear();
    for( int i = 0; i < params->popSize; i ++ ){
        genomes[z][i].eval();
        fitnessStats.add( genomes[z][i].fitness );
    }
}
void Population::sortPopulation(){
    Genome t;
    for( int i = 0; i < params->popSize; i ++ ){
        for( int j = 0; j < params->popSize-1; j ++ ){
            if( genomes[z][j] < genomes[z][j+1] ){
                t = genomes[z][j];
                genomes[z][j] = genomes[z][j+1];
                genomes[z][j+1] = t;
            }
        }
    }
}

void Population::mapFitness( Genome& g ){
    if( fitnessStats.min > 0 )
        g.mappedFitness = g.fitness - fitnessStats.min * .9;
    else
        g.mappedFitness = g.fitness - fitnessStats.min * 1.1;
    //printf("%f -> %f\n", g.fitness, g.mappedFitness );
}

void Population::selection(){
    int i;
    for( i = 0; i < params->elitists; i ++ )
        genomes[1-z][i] = genomes[z][i];
    for( i = params->elitists; i < genomes[z].Size(); i ++ ){
        float choice = drand48() * mappedFitnessStats.sum;
        //printf("choice : %f out of %f\n", choice, mappedFitnessStats.sum );
        for( int j = 0; j < genomes[z].Size(); j ++ ){
            choice -= genomes[z][j].mappedFitness;
            if( choice <= 0 ){
                //printf("genomes[%i][%i] = genomes[%i][%i]\n", 1-z,i, z,j );
                genomes[1-z][i] = genomes[z][j];
                break;
            }
        }
    }
}

void Population::crossover(){
    for( int i = params->elitists; i < genomes[z].Size()-1; i ++ )
        genomes[z][i].crossoverWith( genomes[z][i+1] );
}

void Population::mutation(){
    for( int i = params->elitists; i < genomes[z].Size(); i ++ )
        genomes[z][i].mutate();
}

void Population::logData(){
    FILE* file = fopen("logs/genlogs", "a" );
    fprintf( file, "generation %i done\n", generation );
    fprintf( file, "fitnessstats : %s\n", fitnessStats.toString().Get() );
    fprintf( file, "mappedfitnessstats : %s\n", mappedFitnessStats.toString().Get() );
    fprintf( file, "%s\n", toString());
    fclose( file );
    
    printf("generation %i done\n", generation );
    printf("fitnessstats : %s\n", fitnessStats.toString().Get() );
    printf("mappedfitnessstats : %s\n", mappedFitnessStats.toString().Get() );
    //printf("%s\n", toString());
    genFitnessStats.add( fitnessStats );
}

void Population::logGenStats(){
    FILE* file;
    file = fopen("logs/avglogs", "a" );
    fprintf( file, "%s", genFitnessStats.avg().Get());
    fclose( file );
    file = fopen("logs/minlogs", "a" );
    fprintf( file, "%s", genFitnessStats.min().Get());
    fclose( file );
    file = fopen("logs/maxlogs", "a" );
    fprintf( file, "%s", genFitnessStats.max().Get());
    fclose( file );

    file = fopen("statlogs", "a" );
    nString answer;
    answer += "**** AVGS ****\n" + genFitnessStats.avg();
    answer += "**** MINS ****\n" + genFitnessStats.min();
    answer += "**** MAXS ****\n" + genFitnessStats.max();
    fprintf( file, "%s", answer.Get() );
    fclose( file );
}
